package App::MakeWorld::IORedir;

# ABSTRACT: redirect STDERR, STDOUT

use strict;
use warnings;

our $VERSION = 'v2.5.8';

my ( $ofh, $STDOUT, $STDERR );

# backup real io streams
BEGIN {
    open $STDOUT, '>&', STDOUT;
    open $STDERR, '>&', STDERR;
}

# make a backup of the original filehandles
INIT { $ofh = App::MakeWorld::IORedir->new() }

sub init {

    my $class = shift;

    # first destroy original $ofh.  if this isn't done,
    # it'll get destroyed upon assignment, *after* creating the new
    # one.  this will reset the current fd's to $ofh's saved one
    # *after* the new one has set the, thus confusing things.
    undef $ofh;

    $ofh = $class->new( @_ );

    # now STDOUT and STDERR are set to what is requested, but we also
    # need $ofh to believe those are the values to be used when
    #   MakeWorld::IORedir->new( reset => 1 ) is called
    $ofh->_save;

    return;
}

sub new {

    my ( $class, %attr ) = @_;

    my $self = bless {}, $class;

    $attr{append} //= 1;

    # save current file handles
    $self->_save();


    if ( defined $attr{file} ) {
        open( STDOUT, $attr{append} ? '>>' : '>', $attr{file} )
          or die( "can't redirect STDOUT to $attr{file}\n" );

        open( STDERR, '>&', STDOUT )
          or do {
            open STDOUT, ">&", delete $self->{stdout}
              or die( "error restoring STDOUT\n" );
            die( "can't redirect STDERR to $attr{file}\n" );
          };
    }
    elsif ( defined $attr{reset} ) {
        open STDOUT, ">&", $ofh->{stdout}
          or die( "error resetting STDOUT\n" );

        open STDERR, ">&", $ofh->{stderr}
          or die( "error resetting STDERR\n" );
    }

    elsif ( defined $attr{original} ) {
        open STDOUT, ">&", $STDOUT
          or die( "error resetting STDOUT\n" );

        open STDERR, ">&", $STDERR
          or die( "error resetting STDERR\n" );
    }

    return $self;
}

sub _save {
    my ( $self ) = @_;

    open $self->{stdout}, ">&STDOUT"
      or die( "Can't dup STDOUT: $!\n" );

    open $self->{stderr}, ">&STDERR"
      or die( "Can't dup STDERR: $!\n" );
}


sub DESTROY {
    my $self = shift;

    if ( exists $self->{stdout} ) {
        open STDOUT, ">&", delete $self->{stdout}
          or die( "error restoring STDOUT\n" );
    }

    if ( exists $self->{stderr} ) {
        open STDERR, ">&", delete $self->{stderr}
          or die( "error restoring STDERR\n" );
    }
}

1;

__END__

=pod

=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory

=head1 NAME

App::MakeWorld::IORedir - redirect STDERR, STDOUT

=head1 VERSION

version v2.5.8

=head1 SEE ALSO

Please see those modules/websites for more information related to this module.

=over 4

=item *

L<bin::make-world|bin::make-world>

=back

=head1 AUTHOR

Diab Jerius <djerius@cfa.harvard.edu>

=head1 COPYRIGHT AND LICENSE

This software is Copyright (c) 2018 by Smithsonian Astrophysical Observatory.

This is free software, licensed under:

  The GNU General Public License, Version 3, June 2007

=cut
