package App::MakeWorld::Package;

use strict;
use warnings;

our $VERSION = 'v2.5.8';

use File::Path qw[ rmtree make_path ];
use File::pushd;
use File::Spec::Functions qw[ catdir catfile path file_name_is_absolute ];

use App::MakeWorld::Check qw[ check %BOOLEAN ];

use Cwd qw/ realpath /;

use App::MakeWorld::Global;
use App::MakeWorld::Interpolate;
use App::MakeWorld::Log;
use App::MakeWorld::Run;
use App::MakeWorld::Stages;
use App::MakeWorld::Archive;
use App::MakeWorld::Utils qw[ find_exe find_plugins ];

use namespace::clean;

# environmental variables which are paths.  variables which end in PATH
# are automatically recognized
my %PathVar = map { $_ => 1 } qw[ TEXINPUTS PERLLIB PERL5LIB ];

our @Plugins;
our %PluginMap;

find_plugins( __PACKAGE__, \@Plugins, \%PluginMap );

our @PackageTypes;
our %PackageTypes;

@PackageTypes = grep { !/::(?:Custom|Unknown)$/ } @Plugins;
%PackageTypes = %PluginMap;
delete $PackageTypes{'unknown'};

sub _croak {
    require Carp;
    shift;
    goto \&Carp::croak
}

################################################################################
# create a generic Package object.  This object will later be reblessed into a
# Package subclass once the package type has been discovered.  Nothing can
# be done here which depends upon the package type.

sub new {
    my ( $class, $name, %attr ) = @_;

    my $self = bless {
        name          => $name,
        archive_name  => $attr{archive_name} // $name,
        pkgspec       => defined $attr{pkgspec} ? $attr{pkgspec} : undef,
        tracked_files => [],
        mogrify       => [],
        env           => {},

        # this will be redone once the package is specialized.
        stage => App::MakeWorld::Stages->new(
            stages => App::MakeWorld::Stages::stagedef( $attr{stages} ),
            filter => $Opt->stages,
        ),
    }, $class;

    $self->set_option( $attr{option} ) if exists $attr{option};

    return $self;
}

################################################################################
# Generic initialization of a Package object *after* it has been specialized by
# being reblessed into the appropriate Package subclass.

sub init {
    my $self = shift;

    $self->{stage} = App::MakeWorld::Stages->new(
        stages => $self->can( 'stages' )
        ? $self->stages
        : App::MakeWorld::Stages::stagedef( $self->option( 'stages' ) ),
        filter => $Opt->stages
    );

    $self->{hook}
      = { map { $_ => { pre => [], post => [] } } @{ $self->{stage}{stages} } };

}


################################################################################
# determine if this package class supports the package type based upon the
# name of the archive file

sub matches_archive {
    # my $archive = shift;
    return;
}

################################################################################
# extract the associated archive.  this should be overridden by package classes
# if they expect to return true from the matches_archive() method

sub extract {

    my ( $self ) = @_;

    return $self->archive->extract( $self->logfile( 'extract' ),
        $self->ok_to_run( 'extract' ) );

}

################################################################################
# determine if this package class supports the package type based upon
# the contents of the current directory

sub matches_contents { $_[0]->_croak( "internal error" ); }


################################################################################
# validation template for stages.  package classes may override this

sub stage_vtpl {

#    my ( $self, $stage ) = @_;

    return {
        exe         => { required => 0 },
        arg         => { required => 0 },
        placeholder => { default  => 0 },
    };
}




################################################################################

sub name         { $_[0]->{name} }
sub archive_name { $_[0]->{archive_name} }

sub option {
    my $self = shift;
    @_
      ? ( exists $self->{option}{ $_[0] } ? $self->{option}{ $_[0] } : undef )
      : $self->{option};
}
sub archive { $_[0]->{archive} }
sub base    { $_[0]->{archive}->base }
sub hook    { $_[0]->{hook} }

sub set_option {
    my ( $self, $options ) = @_;

    $self->{option} ||= { %{ $Opt->defoption } };

    for my $hash ( $options, $Opt->option ) {
        $self->{option}{$_} = $hash->{$_} for keys %$hash;
    }

}

sub set_env {

    my $self = shift;

    my %env;
    while ( 'HASH' eq ref $_[0] ) {
        %env = ( %env, %{ shift() } );
    }

    %env = ( %env, @_ );

    $self->{env}{$_} = $env{$_} for keys %env;
    return;

}

################################################################################
# Run the package specification recipe on a generic Package object.  Nothing
# which is specific to a type of Package (i.e. anything with stages) can be
# dealt with here.  It is put off until the process method is called.

sub apply_spec {

    my $self = shift;

    if ( defined $self->{pkgspec} ) {
        my ( $name, $options ) = $self->{pkgspec}->parse( $self );

        $self->_croak( "internal error: expected <package>, got <$name>\n" )
          unless $name eq 'package';

        # delay some checks until after we know what type of package it is
        local $Params::Check::ALLOW_UNKNOWN = 1;

        ## no critic( StringyEval )

        my %tpl = (
            name         => { required => 1 },
            archive_name => { required => 0 },
            pkg_type     => {
                required => 0,
                allow    => [ keys %PackageTypes ]
            },
            matchre => {
                required => 0,
                allow    => [ sub { eval "qr|$_[0]|; 1;" } ],
            },
            versionre => {
                required => 0,
                allow    => [ sub { eval "qr|$_[0]|; 1;" } ],
            },
            version_extract => {
                required => 0,
                allow    => [ sub { eval "qr|$_[0]|; 1;" } ],
            },
            cache => {
                default => defined $Opt->cache,
                %BOOLEAN,
            },

            exclude_if_inpath => { required => 0 },

            exclude => { required => 0, %BOOLEAN },

            libtool => { required => 0 },

            stages => {
                required => 0,
                allow    => [ \&App::MakeWorld::Stages::legal_stage ],
            },

        );

        my $opts = check( \%tpl, $options );

        # remove the options that were validated and save the rest for later
        delete @{$options}{ keys %tpl };
        $self->{options} = $options;

        $self->set_option( $opts );
    }

    return;
}


################################################################################
# validate a package recipe *after* it has been reblessed into the specific
# Package class

sub validate {

    my $self = shift;

    # now that the package type is known, finish validation of recipe
    # these are the options left over from initial validation of the package.
    my $opts = $self->{options};

    # temporary storage for newly validated options
    my %opts;

    # Params::Check is not reentrant, so can't do recursive checks
    for my $stage ( @{ $self->{stage}{stages} } ) {
        if ( defined $opts->{$stage} ) {

            $self->_croak( "<$opts->{name}><$stage>: no attributes specified?\n" )
              unless 'HASH' eq ref $opts->{$stage};

            $opts{$stage}
              = check( $self->stage_vtpl( $stage ), delete $opts->{$stage} );
        }

        else {

            $opts{$stage} = check( $self->stage_vtpl( $stage ), {} );
        }
    }

    $self->set_option( \%opts );

    $self->validate_mogrify_hooks( $_ ) foreach @{ $self->{mogrify} };

    if ( $self->can( 'vtpl' ) ) {
        my $opts = check( $self->vtpl, $self->{options} );
        $self->set_option( $opts );
    }

    return;
}

################################################################################
# set up environment for package

sub setenv {

    my $self = shift;

    # set package specific environment variables. variables with an empty
    # value should be deleted.  this is done twice; once here, and also
    # after merging with $Opt->define and $Opt->odefine.
    # TODO: WHY?
    while ( my ( $var, $value ) = each %{ $self->{env} } ) {
        if ( $value ne '' ) {
            $ENV{$var} = interpolate( $value, $self->archive );
        }
        else {
            delete $ENV{$var};
        }
    }

    # merged environmental variables
    my $env = $self->option( 'env' );
    while ( my ( $var, $value ) = each %{ $Opt->define } ) {
        $value = interpolate( $value, $self->archive );

        if ( $env && exists $env->{$var} ) {
            $value = join( ( $var =~ /PATH$/ || $PathVar{$var} ) ? ':' : ' ',
                $value, $env->{$var} );
        }
        $ENV{$var} = $value;
    }

    # general overrides
    while ( my ( $var, $value ) = each %{ $Opt->odefine } ) {
        $ENV{$var} = interpolate( $value, $self->archive );
    }


    # package specific environmental variable deletions
    delete @ENV{
        map  { $_ }
        grep { $self->{env}{$_} eq '' } keys %{ $self->{env} } };

    return;
}



################################################################################
# find the archive

sub find_archive {
    my $self = shift;
    $self->{archive} = App::MakeWorld::Archive->new( {
            name         => $self->name,
            archive_name => $self->archive_name,
            map { $_ => $self->option( $_ ) }
              qw[ matchre versionre version_extract ],
        },
        @_
    );
}

################################################################################
# process a package.  This:
#
# 1. extracts the package
# 2. determines its type
# 3. finishes validating the recipe
# 4. sets up environment variables
# 5. runs the various stages

sub process {
    my $self = shift;

    my $name = $self->name;

    log_notice( "Building" );

    # if true, check that package contents match package type
    # this is true only for explicitly defined package types
    my $check_contents = 0;

    # switch to the builddir
    my $builddir = catdir( $Opt->builddir, $name );
    make_path( $builddir );

    {
        my $guard = pushd( $builddir );

        # if one of the packages has claimed the archive,
        # extract it and rebless the package into its class
        if ( $self->archive->class ) {
            bless $self, $self->archive->class;
            log_debug( "Assigning to " . $self->archive->class );
        }

        # extract the archive into the builddir
        my $srcdir = $self->extract;

        # switch to the source dir if possible
        if ( -d $srcdir ) {
            chdir( $srcdir )
              or $self->_croak( "unable to cd to $srcdir\n" );

            # if the package class hasn't yet been determined,
            # do it here.
            if ( ref $self eq __PACKAGE__ ) {
                # explicit specification of package type?  normally
                # this is only done for custom types, but we handle 'em all
                # anyway, 'cause we're that sort of folks.
                if ( my $pkg_type = $self->option( 'pkg_type' ) ) {
                    $self->_croak( "unknown package type: $pkg_type\n" )
                      unless exists $PackageTypes{ lc $pkg_type };
                    my $pkgtype = $PackageTypes{ lc $pkg_type };

                    # indicate that we need to verify that the contents are
                    # correct for this package type (if not Custom)
                    $check_contents = $pkgtype !~ /Custom/ ? 1 : 0;

                    bless $self, $pkgtype;
                }

                # determine what type of package we're dealing with.
                else {
                    my ( $pkgclass )
                      = grep { $_->matches_contents } @PackageTypes;

                    $pkgclass //= 'App::MakeWorld::Package::Custom';

                    # if we found something, rebless into that package, else
                    # make it a Custom class and let it complain

                    bless $self, $pkgclass;
                    log_debug( "Assigning to $pkgclass" );
                }
            }

        }

        # no dir and not a dryrun? this shouldn't happen
        elsif ( !$Opt->dryrun ) {
            $self->_croak( "extraction to $srcdir failed?\n" );
        }

        # no srcdir && dryrun; don't know which package it is, so use a
        # do-nothing noisy class
        else {
            App::MakeWorld::Package::Unknown->new( $self );
        }

        $self->init;

        # now that the package type is known, finish validation of recipe
        $self->validate;

        # customize environment
        local %ENV = %ENV;
        $self->setenv;

        # personalize options by interpolating against the archive and
        # command line option values
        {
            my %popt;
            while ( my ( $key, $value ) = each %{ $self->option } ) {
                next if $key eq 'env';
                $popt{$key} = interpolate( $value, $self->archive );
            }
            $self->set_option( \%popt );
        }

        # explicitly do this as extract is a special stage and isn't
        # in the package stage list
        $self->prepost_stage( 'extract', 'post' );

        # check contents against package type if it is required
        $self->_croak( "package does not meet requirements for package type ",
            $self->option( 'pkg_type' ), "\n" )
          if $check_contents && !$self->matches_contents;


        # iterate over stages, performing requested actions
        for my $stage ( @{ $self->{stage}{stages} } ) {
            my $method;

            unless ( $method = $self->can( $stage ) ) {
                log_info(
                    "$stage is not supported for this package class; skipping"
                );
                next;
            }

            $self->prepost_stage( $stage, 'pre' );

            if ( $self->ok_to_run( $stage, 'main' ) ) {
                log_info( $stage );
                eval {
                    my $guard = $self->logredir( $stage );
                    $self->$method;
                };
                if ( my $err = $@ ) {
                    my $logfile = $self->logfile( $stage );
                    $self->_croak( $err,
                        defined $logfile
                        ? "for more info see\n$logfile\n"
                        : () );
                }
            }

            $self->prepost_stage( $stage, 'post' );
        }

    }

    # delete the builddir
    rmtree( $builddir )
      if $Opt->clean
      && !$Opt->dryrun
      && $self->{stage}->ok_to_run( $self->{stage}->last_stage );

    # remove any other files that are tracked
    if ( $Opt->clean && !$Opt->dryrun ) {
        unlink( $_ ) foreach @{ $self->{tracked_files} };
    }
}

################################################################################
# run something before or after a stage

sub prepost_stage {

    my ( $self, $stage, $location ) = @_;

    my @path = ( $stage, $location );

    return unless $self->ok_to_run( @path );

    for my $mogrify ( @{ $self->hook->{$stage}{$location} } ) {
        my $stage = join( ':', @path, $mogrify->{name} );

        log_info( $stage );
        my $guard = log_pfx( $stage );

        # these options haven't yet been interpolated into
        my $opt = interpolate( $mogrify->{attr}, $self->archive );
        eval {
            my $guard = $self->logredir( $stage );
            $mogrify->{class}->mogrify( $self, $opt );
        };
        if ( my $err = $@ ) {
            my $logfile = $self->logfile( $stage );
            $self->_croak( $err,
                defined $logfile ? "for more info see\n$logfile\n" : () );
        }
    }

    return;
}

################################################################################

sub make { main::run( $Opt->make_cmd, @_ ); }


################################################################################

sub ok_to_run {

    my $self = shift;

    return $self->{stage}->ok_to_run( @_ );

}

################################################################################

sub run_stage {

    my ( $self, $stage ) = @_;

    my $option = $self->{option}{$stage};

    return 1 if $option->{placeholder};

    return 0 unless $option->{exe};

    run( find_exe( $option->{exe} ), $option->{arg} )
      or $self->_croak( "error in $stage stage\n" );

    return 1;
}


################################################################################
# it's up to the Package subclasses to perform actions based upon package type
# these are called by those subclass methods to do package specific
# actions.  they return 0 if they haven't handled the action.

sub configure {
    return $_[0]->run_stage( 'configure' );
}

sub build {
    return $_[0]->run_stage( 'build' );
}

sub test {
    return $_[0]->run_stage( 'test' );
}

sub install {
    return $_[0]->run_stage( 'install' );
}

#----------------------------------------------------------------#

sub is_excluded {
    my ( $self ) = @_;

    return 1 if $self->{option}{exclude};

    if ( exists $self->{option}{exclude_if_inpath}{exe} ) {
        return 1
          if $self->find_in_path( $self->{option}{exclude_if_inpath}{exe} );
    }

    return 0;
}

#----------------------------------------------------------------#

sub find_in_path {
    my ( $self, $exe ) = @_;

    my $bindir = realpath( catdir( $Opt->exec_prefix, 'bin' ) );

    foreach my $path ( path() ) {
        my $rpath = realpath( $path );

        next if !defined $rpath or $rpath eq $bindir;

        return 1 if -x catfile( $path, $exe );
    }

    return;
}

#----------------------------------------------------------------#




#----------------------------------------------------------------#

sub logfile {
    my ( $self, $stage ) = @_;

    my $logfile;

    if ( $Opt->logredir ) {
        $logfile = catfile( $Opt->builddir, $self->name, "$stage.log" );
        push @{ $self->{tracked_files} }, $logfile;
    }


    return $logfile;
}

sub logredir {
    my ( $self, $stage ) = @_;

    return App::MakeWorld::IORedir->new(
        file   => $self->logfile( $stage ),
        append => 0
    );
}


################################################################################
#
# mogrification interface

# a package may provide transformation routines which will act upon the
# package directory at the various stages.  This is a somewhat kludgy means
# of providing verification and running of those stages

################################################################################
# called by the recipe parser to add a mogrification command to the
# package's list.

sub add_mogrify {

    my ( $self, $mogrify ) = @_;

    push @{ $self->{mogrify} }, $mogrify;

    return;
}

################################################################################
# validate that the requested hook is legit and add the command to the
# hook list.  called after a package has been specialized

sub validate_mogrify_hooks {

    my $self    = shift;
    my $mogrify = shift;

    $self->_croak( "invalid hook attribute value for $mogrify->{name}: $mogrify->{hook}\n"
    ) unless $self->{stage}->is_legal_hook( $mogrify->{hook} );

    my ( $stage, $location ) = $self->{stage}->parse_hook( $mogrify->{hook} );

    push @{ $self->{hook}{$stage}{$location} }, $mogrify;

}

1;

__END__

=pod

=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory

=head1 NAME

App::MakeWorld::Package

=head1 VERSION

version v2.5.8

=head1 SEE ALSO

Please see those modules/websites for more information related to this module.

=over 4

=item *

L<bin::make-world|bin::make-world>

=back

=head1 AUTHOR

Diab Jerius <djerius@cfa.harvard.edu>

=head1 COPYRIGHT AND LICENSE

This software is Copyright (c) 2018 by Smithsonian Astrophysical Observatory.

This is free software, licensed under:

  The GNU General Public License, Version 3, June 2007

=cut
