package App::MakeWorld::Package::LuaRock;

use strict;
use warnings;

our $VERSION = 'v2.5.8';

use File::Basename;
use File::Spec::Functions qw/ catdir catfile /;
use File::Temp qw( tempdir );
use File::Copy;
use File::Path;
use Try::Tiny;
use Cwd qw( getcwd );

use App::MakeWorld::Global;
use App::MakeWorld::Log;
use App::MakeWorld::Run;
use App::MakeWorld::Utils qw[ find_exe ];


use base 'App::MakeWorld::Package';

use namespace::clean;

sub stages { [qw[ install ]] }

sub vtpl {

    return {
        subdir => { required => 0 },
        arg    => { required => 0 },
    };

}

sub matches_contents { 0; }

sub matches_archive {

    my ( $class, $archive ) = @_;

    my $ext = qr /\.src\.rock$/;
    return unless $archive =~ $ext;

    return fileparse( $archive, $ext );

}

sub extract_version {

    my ( $class, $version ) = @_;

    my @version = $version =~ /^
                              (\d+)
                              (?:\.(\d+))?
                              (?:\.(\d)?)?
                              (\w+)?
                              -
                              (\d+)
                              $/x;

    $class->die( "unable to extract version information from $version\n" )
      unless @version;

    return join( '.', map { $_ // 0 } @version[ 0, 1, 2 ] ),
      $version[3], $version[4];
}

sub extract {
    my ( $self ) = @_;

    my $base = $self->archive->{base};
    my $file = $self->archive->{file};

    my $cwd = getcwd;

    $self->{luarocks} = find_exe( 'luarocks' );
    # directory where the archive is extracted
    my $extdir;


    # this is the name of the directory that'll get renamed as $base.
    # either $extdir or $extdir/subdir
    my $srcdir;

    if ( $self->ok_to_run( 'extract' ) ) {

        try {
            # extract archive into its own directory.  this allows
            # processing of archives which don't unpack into
            # a properly named and versioned directory

            unless ( $Opt->dryrun ) {
                $extdir = tempdir( DIR => $cwd );
                chdir( $extdir )
                  or $self->_croak( "unable to chdir to $extdir\n" );
            }

            log_info( "extracting $file" );


            eval {
                my $guard = $self->logredir( 'extract' );

                # luarocks tries to optimize determination of the current
                # working directory by using the PWD environment variable,
                # but that's not correct.  remove the temptation...
                local %ENV = %ENV;
                delete $ENV{PWD};
                run( $self->{luarocks}, 'unpack', $file )
                  or $self->_croak( "error extracting package to $extdir\n" );
            };
            if ( my $err = $@ ) {
                my $logfile = $self->logfile( 'extract' );
                $self->_croak( $err,
                    defined $logfile ? "for more info see\n$logfile\n" : () );
            }

            if ( !$Opt->dryrun ) {
                opendir( my $dh, '.' )
                  or $self->_croak( 'unable to open current directory?' );
                # now, what's in the current directory?
                my @dircontents = grep { !/^\./ } readdir( $dh );
                closedir( $dh );
                # only one thing and its a directory. cool.
                if ( 1 == @dircontents && -d $dircontents[0] ) {
                    $srcdir = catdir( $extdir, $dircontents[0] );
                }

                # more than one entry. rename $extdir to $base
                else {
                    $self->_croak( "extracted more than expected: got ",
                        join( ", ", @dircontents ), "\n" );
                }
            }

            chdir( $cwd );
            # remove build directory if it exists
            if ( -d $base ) {
                log_info( "removing existing build in $cwd/$base" );
                rmtree( $base )
                  unless $Opt->dryrun;
            }

            unless ( $Opt->dryrun ) {
                move( $srcdir, $base )
                  or $self->_croak( "error renaming $srcdir to $base\n" );
            }

        }

        catch {
            $self->_croak($_);
        }

        finally {
            chdir( $cwd );
            # remove extdir if it hasn't been renamed to $base
            rmtree( $extdir ) if defined $extdir && -d $extdir;
        };

    }

    # now that it's extracted, find the rockspec and the actual directory
    # containing the code
    return $self->_find_srcdir( $cwd, $base );

}

sub _find_srcdir {

    my ( $self, $cwd, $base ) = @_;

    my $srcdir = catdir( $cwd, $base );

    chdir( $srcdir )
      or $self->_croak( "unable to change directory to $srcdir\n" );

    my ( @rockspec ) = glob( '*.rockspec' );

    $self->_croak( "expected one rockspec file, got ", scalar @rockspec, "\n" )
      unless 1 == @rockspec;

    $self->{rockspec} = catfile( $srcdir, $rockspec[0] );

    my $subdir = $self->option( 'subdir' );

    if ( !defined $subdir ) {
        opendir( my $dh, '.' )
          or $self->_croak( 'unable to open current directory?' );
        my @dirs = grep { !/^\./ && -d $_ } readdir( $dh );
        closedir( $dh );

        $self->_croak( "can't figure out which directory contains the code\n" )
          if @dirs > 1;

        $subdir = $dirs[0];
    }

    return catdir( $cwd, $base, $subdir );
}


sub install {
    my $self = shift;

    $self->SUPER::build && return;

    # luarocks tries to optimize determination of the current
    # working directory by using the PWD environment variable,
    # but that's not correct.  remove the temptation...
    local %ENV = %ENV;
    delete $ENV{PWD};

    run( $self->{luarocks}, $self->option( 'arg' ), 'make', $self->{rockspec} )
      or $self->_croak( "error building and installing package\n" );

    return;
}


1;

__END__

=pod

=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory

=head1 NAME

App::MakeWorld::Package::LuaRock

=head1 VERSION

version v2.5.8

=head1 SEE ALSO

Please see those modules/websites for more information related to this module.

=over 4

=item *

L<bin::make-world|bin::make-world>

=back

=head1 AUTHOR

Diab Jerius <djerius@cfa.harvard.edu>

=head1 COPYRIGHT AND LICENSE

This software is Copyright (c) 2018 by Smithsonian Astrophysical Observatory.

This is free software, licensed under:

  The GNU General Public License, Version 3, June 2007

=cut
