package App::MakeWorld::Run;

use strict;
use warnings;

our $VERSION = 'v2.5.8';

use File::Basename;

use App::MakeWorld::Global;
use App::MakeWorld::Log;

use Capture::Tiny ();

use base qw[ Exporter ];

our @EXPORT = qw[
  run
  capture
];

#----------------------------------------------------------------#

# switches for programs which can handle "dry run" flags
our %dryrun = (
    make      => '-n',
    gmake     => '-n',
    configure => '-n',
);

sub _run_setup {
    my ( $exe, @args ) = @_;

    # @args may have an arrayref hidden in there, so take care of it
    @args = map { 'ARRAY' eq ref $_ ? @$_ : $_ } grep { defined $_ } @args;

    my $lexe = basename( $exe );

    unshift @args, $dryrun{$lexe}
      if $Opt->dryrun > 1 && exists $dryrun{$lexe};

    my $cmd = join( ' ', $exe, @args );

    log_at_level(
        $Opt->dryrun ? App::MakeWorld::Log::Notice : App::MakeWorld::Log::Info,
        "executing $cmd"
    );

    return ( $exe, $lexe, @args );
}

# wrapper for system() command which knows about dryruns
sub run {
    my ( $exe, $lexe, @args ) = _run_setup( @_ );

    # this level of dryrun means just print what we'd do
    return 1 if 1 == $Opt->dryrun;

    # if a command can do a dryrun, have it do so, otherwise
    # return
    return 1 if $Opt->dryrun && !exists $dryrun{$lexe};

    return system( $exe, @args ) == 0;
}


# wrapper for system() command which knows about dryruns
sub capture {
    my ( $exe, $lexe, @args ) = _run_setup( @_ );

    # this level of dryrun means just print what we'd do
    return ( 0 ) if 1 == $Opt->dryrun;

    # if a command can do a dryrun, have it do so, otherwise
    # return
    return ( 0 ) if $Opt->dryrun && !exists $dryrun{$lexe};

    my $exit;
    my ( $stdout, $stderr )
      = Capture::Tiny::capture { $exit = system( $exe, @args ) };

    return ( $exit, $stdout, $stderr );
}


1;

__END__

=pod

=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory

=head1 NAME

App::MakeWorld::Run

=head1 VERSION

version v2.5.8

=head1 SEE ALSO

Please see those modules/websites for more information related to this module.

=over 4

=item *

L<bin::make-world|bin::make-world>

=back

=head1 AUTHOR

Diab Jerius <djerius@cfa.harvard.edu>

=head1 COPYRIGHT AND LICENSE

This software is Copyright (c) 2018 by Smithsonian Astrophysical Observatory.

This is free software, licensed under:

  The GNU General Public License, Version 3, June 2007

=cut
